/*	SpacesInjector.m
	SpacesSpacesSpaces is licensed under the BSD-license:
	
	Copyright (c) 2007, Daniel Stoedle <daniels@scsc.no>,
	Yellow Lemon Software. All rights reserved.
	
	Redistribution and use in source and binary forms, with or without
	modification, are permitted provided that the following conditions are met:

	- Redistributions of source code must retain the above copyright notice,
	  this list of conditions and the following disclaimer.

	- Redistributions in binary form must reproduce the above copyright notice,
	  this list of conditions and the following disclaimer in the documentation
	  and/or other materials provided with the distribution.

	- Neither the name of Yellow Lemon Software nor the names of its
	  contributors may be used to endorse or promote products derived from this
	  software without specific prior written permission.

	THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
	AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
	IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
	ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
	LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
	CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
	SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
	INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
	CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
	ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
	POSSIBILITY OF SUCH DAMAGE.
*/

#import		"SpacesInjector.h"
#import		"mach_inject_bundle.h"
#include	<mach-o/dyld.h>
#include	<Security/Authorization.h>
#include	<Security/AuthorizationTags.h>

@implementation SpacesInjector
- (void)awakeFromNib {
}


- (void)updateDockPID {
	size_t		proc_count = 0, i;
	kinfo_proc	*procs	= 0;
	
	if (dockPid)
		[dockPid release];
	dockPid	= 0;
	GetBSDProcessList(&procs, &proc_count);
	for (i=0;i<proc_count;i++) {
		if (procs[i].kp_eproc.e_pcred.p_ruid == getuid()) {
			if (strcmp(procs[i].kp_proc.p_comm, "Dock") == 0) {
				dockPid	= [[NSString alloc] initWithFormat:@"%d", procs[i].kp_proc.p_pid];
				break;
			}
		}
	}
	//	We don't care about leaking the resources used by the process list.
	//	This is not supposed to be a long-lived application.
	//NSLog(@"Dock: %@ PID: %@", [[NSWorkspace sharedWorkspace] fullPathForApplication:@"Dock"], dockPid);
}

- (BOOL)dockIsCompatible {
	pid_t	pid;
	
	pid	= [dockPid intValue];
	if (pid > 0) {
		ProcessSerialNumber	psn;
		
		if (GetProcessForPID(pid, &psn) == noErr) {
			NSDictionary	*dict;
			NSString		*bundlePath;
			
			dict		= (NSDictionary*)ProcessInformationCopyDictionary(&psn, kProcessDictionaryIncludeAllInformationMask);
			bundlePath	= [dict objectForKey:@"BundlePath"];
			if (bundlePath) {
				NSBundle	*bundle;
				NSString	*version;
				
				bundle	= [NSBundle bundleWithPath:bundlePath];
				dict	= [bundle infoDictionary];
				version	= [dict objectForKey:@"CFBundleVersion"];
				if (version) {
					if ([version compare:@"578"] == NSOrderedSame)
						return YES;
				}
			}
		}
	}
	return NO;
}


- (IBAction)inject:(id)sender {
	NSString			*path	= [[NSBundle mainBundle] pathForResource:@"SpacesInjector" ofType:@""];
	OSStatus			err;
	AuthorizationFlags	flags	= kAuthorizationFlagDefaults;
	AuthorizationRef	auth_ref;
	const char			*args[]	= {	[[[NSBundle mainBundle] pathForResource:@"InjectStage1" ofType:@"bundle"] UTF8String],
									0,
									0 };
	
	[self updateDockPID];
	if (dockPid) {
		//	Check that the Dock is compatible; this hack is only tested with v. 578
		if ([self dockIsCompatible]) {
			args[1]	= [dockPid UTF8String];
			if (geteuid() != 0 && getuid() != 0) {
				int	result = NSRunAlertPanel(@"Spaces.. Spaces.. Spaces.. must run as root", @"Spaces.. Spaces.. Spaces.. must run as the administrator in order to inject its own code into the Dock.", @"Authenticate", @"Do nothing", 0);
				if (result == NSOKButton) {
					auth_ref	= [self authorizeUser];
					if (!auth_ref)
						return;
				}
				else
					return;
			}
			err = AuthorizationExecuteWithPrivileges(auth_ref, [path UTF8String], flags, args, NULL);
			AuthorizationFree(auth_ref, kAuthorizationFlagDefaults);
			if (err == errAuthorizationSuccess)
				NSRunAlertPanel(@"Success!", @"The code has been successfully injected. Try switching between applications to see if it worked.",  @"Great!",  0, 0);
			else
				NSRunAlertPanel(@"Error!", [NSString stringWithFormat:@"The code failed inject properly. The error code is %d.", err],  @"Bummer",  0, 0);
		}
		else
			NSRunAlertPanel(@"Error!", @"Your version of the Dock is either to old or too new.", @"OK",  0, 0);
	}
	else
		NSRunAlertPanel(@"Error!", @"Failed to get the Dock's process ID.", @"OK",  0, 0);
}


- (IBAction)relaunchDock:(id)sender {
	NSAppleScript *script	= [[NSAppleScript alloc] initWithSource:@"tell app \"Dock\" to quit"];
	[script executeAndReturnError:0];
}


- (IBAction)sendFeedback:(id)sender {
	[[NSWorkspace sharedWorkspace] openURL:[NSURL URLWithString:@"mailto:yls@scsc.no?subject=Spaces..%20Spaces..%20Spaces.."]];
}


- (IBAction)visitWebsite:(id)sender {
	[[NSWorkspace sharedWorkspace] openURL:[NSURL URLWithString:@"http://www.scsc.no/products/spaces-spaces-spaces/"]];
}


- (IBAction)donate:(id)sender {
	[[NSWorkspace sharedWorkspace] openURL:[NSURL URLWithString:@"http://order.kagi.com/?W8"]];
}


- (AuthorizationRef)authorizeUser {
	OSStatus			err;
	AuthorizationFlags	flags	= kAuthorizationFlagDefaults;
	AuthorizationRef	auth_ref;
	AuthorizationItem	items	= {kAuthorizationRightExecute, 0, NULL, 0};
	AuthorizationRights	rights	= {1, &items};
	
	err = AuthorizationCreate(NULL, kAuthorizationEmptyEnvironment, flags, &auth_ref);
	if (err != errAuthorizationSuccess)
		return 0;
	
	flags	= kAuthorizationFlagDefaults | kAuthorizationFlagInteractionAllowed | kAuthorizationFlagExtendRights | kAuthorizationFlagPreAuthorize;
	err		= AuthorizationCopyRights(auth_ref, &rights, NULL, flags, NULL);
	if (err != errAuthorizationSuccess) {
		AuthorizationFree(auth_ref, kAuthorizationFlagDefaults);
		return 0;
	}
	
	return auth_ref;
}
@end

//	This is sample code from Apple. I can't believe there isn't a simpler solution
//	to getting the Dock's pid! (And no, [NSWorkspace launchedApplications] isn't it)
//	http://developer.apple.com/qa/qa2001/qa1123.html


int GetBSDProcessList(kinfo_proc **procList, size_t *procCount)
    // Returns a list of all BSD processes on the system.  This routine
    // allocates the list and puts it in *procList and a count of the
    // number of entries in *procCount.  You are responsible for freeing
    // this list (use "free" from System framework).
    // On success, the function returns 0.
    // On error, the function returns a BSD errno value.
{
    int                 err;
    kinfo_proc *        result;
    bool                done;
    static const int    name[] = { CTL_KERN, KERN_PROC, KERN_PROC_ALL, 0 };
    // Declaring name as const requires us to cast it when passing it to
    // sysctl because the prototype doesn't include the const modifier.
    size_t              length;

	if (procList == 0)
		return 1;
	if (*procList != 0)
		return 1;
	if (procCount == 0)
		return 1;
		
    *procCount = 0;

    // We start by calling sysctl with result == NULL and length == 0.
    // That will succeed, and set length to the appropriate length.
    // We then allocate a buffer of that size and call sysctl again
    // with that buffer.  If that succeeds, we're done.  If that fails
    // with ENOMEM, we have to throw away our buffer and loop.  Note
    // that the loop causes use to call sysctl with NULL again; this
    // is necessary because the ENOMEM failure case sets length to
    // the amount of data returned, not the amount of data that
    // could have been returned.

    result = NULL;
    done = false;
    do {
        if (result != 0)
			break;

        // Call sysctl with a NULL buffer.

        length = 0;
        err = sysctl( (int *) name, (sizeof(name) / sizeof(*name)) - 1,
                      NULL, &length,
                      NULL, 0);
        if (err == -1) {
            err = errno;
        }

        // Allocate an appropriately sized buffer based on the results
        // from the previous call.

        if (err == 0) {
            result = malloc(length);
            if (result == NULL) {
                err = ENOMEM;
            }
        }

        // Call sysctl again with the new buffer.  If we get an ENOMEM
        // error, toss away our buffer and start again.

        if (err == 0) {
            err = sysctl( (int *) name, (sizeof(name) / sizeof(*name)) - 1,
                          result, &length,
                          NULL, 0);
            if (err == -1) {
                err = errno;
            }
            if (err == 0) {
                done = true;
            } else if (err == ENOMEM) {
				if (result == 0)
					break;
                free(result);
                result = NULL;
                err = 0;
            }
        }
    } while (err == 0 && ! done);

    // Clean up and establish post conditions.

    if (err != 0 && result != NULL) {
        free(result);
        result = NULL;
    }
    *procList = result;
    if (err == 0) {
        *procCount = length / sizeof(kinfo_proc);
    }

    return err;
}
