/*	TripleS.m
	SpacesSpacesSpaces is licensed under the BSD-license:
	
	Copyright (c) 2007, Daniel Stoedle <daniels@scsc.no>,
	Yellow Lemon Software. All rights reserved.
	
	Redistribution and use in source and binary forms, with or without
	modification, are permitted provided that the following conditions are met:

	- Redistributions of source code must retain the above copyright notice,
	  this list of conditions and the following disclaimer.

	- Redistributions in binary form must reproduce the above copyright notice,
	  this list of conditions and the following disclaimer in the documentation
	  and/or other materials provided with the distribution.

	- Neither the name of Yellow Lemon Software nor the names of its
	  contributors may be used to endorse or promote products derived from this
	  software without specific prior written permission.

	THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
	AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
	IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
	ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
	LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
	CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
	SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
	INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
	CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
	ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
	POSSIBILITY OF SUCH DAMAGE.
*/

#import "TripleS.h"
#include "mach_override.h"

static CPSSetFrontProcessWithOptionsFunc	original_CPSSetFrontProcessWithOptions = 0;
static DockSwitchWorkspaceThreadFunc		original_dockSwitchWorkspace = 0;
static double								last_set_front_process = 0.;

@implementation TripleS
+ (TripleS*)sharedInstance {
	static TripleS	*sss = 0;
	if (sss == 0)
		sss = [[TripleS alloc] init];
	return sss;
}

+ (void)overrideWorkspaceSwitch {
	NSDictionary		*dict;
	NSString			*version;
	NSAutoreleasePool	*pool	= [[NSAutoreleasePool alloc] init];
	
	dict	= [[NSBundle mainBundle] infoDictionary];
	version	= [dict objectForKey:@"CFBundleVersion"];
	if (version) {
		if ([version compare:@"578"] == NSOrderedSame) {
			if (mach_override("__CPSSetFrontProcessWithOptions", NULL, TripleS_CPSSetFrontProcessWithOptions, (void**) &original_CPSSetFrontProcessWithOptions) != 0) {
				printf("TripleS: Override failed: CPSSetFrontProcessWithOptions\n");
				return;
			}
			//	It took some reverse engineering to figure out the address hardcoded below.
			//	TODO: This address must be changed to work on PowerPC.
			if (mach_override_ptr((void*)0x80445, TripleS_DockSwitchWorkspace, (void**) &original_dockSwitchWorkspace) != 0) {
				printf("TripleS: Override-ptr failed: 0x80445\n");
				return;
			}
			printf("TripleS: Override successful!\n");
			return;
		}
		else
			printf("TripleS: Override failed. Incompatible Dock version: %s (should be 578)\n", [version UTF8String]);
	}
	else
		printf("TripleS: Override failed. Failed to get Dock version\n");
	[pool release];
}
@end

OSStatus	TripleS_CPSSetFrontProcessWithOptions(CPSProcessSerNum *psn, OptionBits inOptions) {
	last_set_front_process	= double_time();
	//printf("TripleS: Set CPS front process, options %x tstamp = %f\n", inOptions, last_set_front_process);
	if (original_CPSSetFrontProcessWithOptions)
		return original_CPSSetFrontProcessWithOptions(psn, inOptions);
	return 0;
}


void*		TripleS_DockSwitchWorkspace(void *arg) {
	double	 now = double_time();
	if (now - last_set_front_process < 0.2) {
		//printf("TripleS: Inhibit desktop switch: %f\n", now - last_set_front_process);
		return 0;
	}
	//printf("TripleS: DockSwitchWorkspace.\n");
	if (original_dockSwitchWorkspace)
		return original_dockSwitchWorkspace(arg);
	return 0;
}


double		double_time(void) {
	struct timeval now;
	
	gettimeofday(&now, 0);
	return (double)now.tv_sec + (((double)now.tv_usec)/10e5f);
}
